#include "mpi.h"
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <sys/time.h>

#define SEED 13517103
#define MAX_VALUE -1

void dijkstra(int *G, int n, int startnode, long *distance);

int main(int argc, char *argv[]) {
  int world_size, rank;
  int N, i, j;
  long long total_time;
  char file_name[] = "output.txt";
  FILE * fp;

  N = atoi(argv[1]);
  
//   n_proc = atoi(argv[2]);

  fp = fopen(file_name,"w");
  
  int *G = (int *)malloc (N * N * sizeof(int));

  if (G==NULL) {
    fprintf(stderr, "Usage: memory allocation failed\n");
    exit(1);
  }

  int *local_size = (int *)malloc (sizeof(int));

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &world_size);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);

  if (rank == 0) {
    srand(SEED);

    // Pembangkitan graf
    for (i=0; i<N; i++) {
      for (j=0; j<N; j++) {
        if (i==j)     G[i * N + j] = MAX_VALUE;
        else if (i>j) G[i * N + j] = G[j * N + i];
        else          G[i * N + j] = rand();
      }
    }
  }

  // MPI_Barrier(MPI_COMM_WORLD);
  // total_time -= MPI_Wtime();
  struct timeval st, et;
  if(!rank) gettimeofday(&st, NULL);

  MPI_Bcast(G, N*N, MPI_INT, 0, MPI_COMM_WORLD);
  *local_size = N / world_size;
  MPI_Bcast(local_size, 1, MPI_INT, 0, MPI_COMM_WORLD);

  long *temp = (long*) malloc ((*local_size) * N * sizeof(long));
  if (temp==NULL) {
    fprintf(stderr, "Usage: memory allocation failed\n");
    exit(1);
  }

  int ctr = 0;

  // Dijkstra here
  for (i=(rank* (*local_size)); i<(rank* (*local_size)+(*local_size)); i++) {
    dijkstra(G, N, i, temp+(ctr * N));
    ++ctr;
  }

  long *result = NULL;
  
  if (rank==0) {
    result = (long *)malloc (N * N * sizeof(long));
    if (result==NULL) {
      fprintf(stderr, "Usage: memory allocation failed\n");
      exit(1);
    }
    
    // Result
    for (i=0; i<N*N; i++) {
      result[i] = 0;
    }
  }

  MPI_Gather(temp, (*local_size)*N, MPI_LONG, result, (*local_size)*N, MPI_LONG, 0, MPI_COMM_WORLD);

  // MPI_Barrier(MPI_COMM_WORLD);
  // total_time += MPI_Wtime();
  if (!rank) {
    gettimeofday(&et, NULL);
    total_time = ((et.tv_sec - st.tv_sec) * 1000000) + (et.tv_usec - st.tv_usec);
  }

  if(!rank) {
    for (i=0; i<N; i++) {
      for(j=0; j<N; j++) {
        fprintf(fp, "%li ", result[i * N + j]);
      }
      fprintf(fp, "\n");
    }
  }
  

  fclose(fp);
  free(temp);
  free(G);

  if(!rank) free(result);

  MPI_Finalize();

  printf("%lli", total_time);

  return 0;
}

void dijkstra(int *G, int n, int startnode, long *distance) {
 int *pred, *visited;
 int count, next_node, i, j;
 long min_distance;

  pred = (int *) malloc(n*sizeof(int));
  if (pred==NULL) {
    fprintf(stderr, "Usage: memory allocation failed\n");
    exit(1);
  }

  visited = (int *) malloc(n*sizeof(int));
  if (visited==NULL) {
    fprintf(stderr, "Usage: memory allocation failed\n");
    exit(1);
  }
 
 // pred[] stores the predecessor of each node
 // count gives the number of nodes seen so far
 
 // initialize pred[],distance[] and visited[]
 for (i=0; i<n; i++) {
  distance[i] = G[startnode * n + i];
  pred[i] = startnode;
  visited[i] = 0;
 }
 
 distance[startnode] = 0;
 visited[startnode] = 1;
 count = 1;

 
 while (count < (n-1)) {
  min_distance = LONG_MAX;
  
  // next_node gives the node at minimum distance
  for (i=0; i<n; i++) {
      if (distance[i]==MAX_VALUE) continue;
   else if (distance[i]<min_distance && !visited[i]) {
    min_distance = distance[i];
    next_node = i;
      }
    }
        
    // check if a better path exists through next_node   
    visited[next_node] = 1;
    for (i=0; i<n; i++) {
      if (!visited[i]) {
        if (distance[i]==MAX_VALUE && G[next_node * n + i]!=MAX_VALUE) {
          distance[i] = min_distance + G[next_node * n + i];
          pred[i] = next_node;
        }
        else if (min_distance+G[next_node * n + i]<distance[i] && G[next_node * n + i]!=MAX_VALUE) {
          distance[i] = min_distance + G[next_node * n + i];
          pred[i] = next_node;
        }
      }
    }
  count++;
 }

  free(pred);
  free(visited);
}